/* -*-c-*-
 *
 * Pyke: the Python Kit for Extensions, mLib integration
 *
 * (c) 2019 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Pyke: the Python Kit for Extensions.
 *
 * Pyke is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Pyke is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Pyke.  If not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include "pyke-mLib.h"

/* #undef HAVE_LONG_LONG */

/*----- Conversions -------------------------------------------------------*/

#ifndef HAVE_LONG_LONG
static PyObject *i32 = 0;
static int init_i32(void)
  { if (!i32 && (i32 = PyInt_FromLong(32)) == 0) return (-1); return (0); }
#endif

PyObject *getk64(kludge64 u)
{
#ifdef HAVE_LONG_LONG
  return (PyLong_FromUnsignedLongLong(GET64(unsigned PY_LONG_LONG, u)));
#else
  PyObject *i = 0, *j = 0, *t;
  PyObject *rc = 0;

  if (init_i32()) goto end;
  if ((i = PyLong_FromUnsignedLong(HI64(u))) == 0) goto end;
  if ((t = PyNumber_InPlaceLshift(i, i32)) == 0) goto end;
  Py_DECREF(i); i = t;
  if ((j = PyLong_FromUnsignedLong(LO64(u))) == 0) goto end;
  if ((t = PyNumber_InPlaceOr(i, j)) == 0) goto end;
  Py_DECREF(i); i = t;
  if ((t = PyNumber_Int(i)) == 0) goto end;
  rc = t;
end:
  Py_XDECREF(i);
  Py_XDECREF(j);
  return (rc);
#endif
}

#ifdef HAVE_UINT64
#  define CONVu64(n) do {						\
     kludge64 k;							\
     uint64 t;								\
     if (!convk64(o, &k)) goto end;					\
     t = GET64(uint64, k);						\
     if (t > MASK##n) VALERR("out of range");				\
     *p = t;								\
   } while (0)
#else
#  define CONVu64(n) assert(!"shouldn't be possible")
#endif

#define CONVU_(n)							\
  int convu##n(PyObject *o, void *pp)					\
  {									\
    unsigned long u;							\
    uint##n *p = pp;							\
									\
    if (MASK##n > ULONG_MAX)						\
      CONVu64(n);							\
    else {								\
      if (!convulong(o, &u)) goto end;					\
      if (u > MASK##n) VALERR("out of range");				\
      *p = u;								\
    }									\
    return (1);								\
  end:									\
    return (0);								\
  }
DOUINTSZ(CONVU_)

int convk64(PyObject *o, void *pp)
{
  PyObject *i = 0;
  int rc = 0;
#if HAVE_LONG_LONG
  unsigned PY_LONG_LONG t;
#else
  PyObject *t;
  uint32 lo, hi;
#endif

#if HAVE_LONG_LONG
  if ((i = PyNumber_Long(o)) == 0) goto end;
  t = PyLong_AsUnsignedLongLong(i);
  if (t == (unsigned PY_LONG_LONG)-1 && PyErr_Occurred()) goto end;
  ASSIGN64(*(kludge64 *)pp, t);
#else
  if (init_i32()) goto end;
  if ((i = PyNumber_Int(o)) == 0) goto end;
  lo = PyInt_AsUnsignedLongMask(i);
  if ((t = PyNumber_InPlaceRshift(i, i32)) == 0) goto end;
  Py_DECREF(i); i = t;
  hi = PyInt_AsUnsignedLongMask(i);
  if ((t = PyNumber_InPlaceRshift(i, i32)) == 0) goto end;
  Py_DECREF(i); i = t;
  if (PyObject_IsTrue(i)) VALERR("out of range");
  SET64(*(kludge64 *)pp, hi, lo);
#endif
  rc = 1;

end:
  Py_XDECREF(i);
  return (rc);
}

/*----- That's all, folks -------------------------------------------------*/
